@@ -34,18 +34,26 @@ class BKTreeNode[ValueT]:
         if len(self.children.items()) != len(other.children.items()):
             return False
 
-        for (distance_1, child_1), (distance_2, child_2) in zip(self.children.items(), other.children.items()):
+        for (distance_1, child_1), (distance_2, child_2) in zip(
+            self.children.items(), other.children.items()
+        ):
             if distance_1 != distance_2 or child_1 != child_2:
                 return False
 
         return True
 
     @classmethod
-    def from_dict(cls, node_input, value_type: type[ValueT] = str):
+    def from_dict(
+        cls,
+        node_input,
+        value_type: type[ValueT] = str,
+    ):
         if value_type is str:
             new_node = cls(value=node_input["value"], value_type=value_type)
         else:
-            new_node = cls(value=value_type(**node_input["value"]), value_type=value_type)
+            new_node = cls(
+                value=value_type(**node_input["value"]), value_type=value_type
+            )
         for key, value in node_input["children"].items():
             new_node.children[int(key)] = cls.from_dict(value)
         return new_node
@@ -66,28 +74,35 @@ class BKTreeNode[ValueT]:
         return {
             "value": self.value,
             "children": {
-                str(distance): child_node.to_dict() for distance, child_node in self.children.items()
-            }
+                str(distance): child_node.to_dict()
+                for distance, child_node in self.children.items()
+            },
         }
 
 
-class BKTree[ValueT, NodeT]:
-
+class BKTree[ValueT, NodeT: BKTreeNode]:
     def __init__(
-            self,
-            distance_function: Callable[[Sequence[Hashable], Sequence[Hashable]], int],
-            max_distance: int = 0,
-            key: Callable[[ValueT], str] | None = None,
-            value_type: type[ValueT] = str,
-            node_type: type[NodeT] = BKTreeNode,
-            *,
-            post_add_hook: Callable[[Self, NodeT], None] = lambda tree, node: None,
-            post_already_added_hook: Callable[[Self, NodeT], None] = lambda tree, node: None,
-            post_filter_add_hook: Callable[[Self, NodeT], None] = lambda tree, node: None,
-            post_insert_subtree_hook: Callable[[Self, NodeT], None] = lambda tree, node: None,
-            post_combine_hook: Callable[[Self, NodeT], None] = lambda tree, node: None
+        self,
+        distance_function: Callable[[Sequence[Hashable], Sequence[Hashable]], int],
+        max_distance: int = 0,
+        key: Callable[[ValueT], str] | None = None,
+        value_type: type[ValueT] = str,
+        node_type: type[NodeT] = BKTreeNode,
+        *,
+        post_add_hook: Callable[[Self, NodeT], None] = lambda tree, node: None,
+        post_already_added_hook: Callable[[Self, NodeT], None] = lambda tree,
+        node: None,
+        post_filter_add_hook: Callable[[Self, NodeT, NodeT], None] = lambda tree,
+        node,
+        filtered_node: None,
+        post_insert_subtree_hook: Callable[[Self, NodeT, NodeT], None] = lambda tree,
+        node,
+        subtree_root: None,
+        post_combine_hook: Callable[[Self, NodeT, NodeT], None] = lambda tree,
+        node,
+        combined_node: None,
     ):
-        self.root: BKTreeNode[ValueT] | None = None
+        self.root: NodeT | None = None
         self.distance_function = distance_function
         self.max_distance = max_distance
         self.nb_elements = 0
@@ -101,28 +116,96 @@ class BKTree[ValueT, NodeT]:
         self.post_combine_hook = post_combine_hook
 
     @classmethod
-    def from_json(cls, filepath: os.PathLike | str, distance_function: Callable[[Sequence[Hashable], Sequence[Hashable]], int],
-                  key: Callable[[ValueT], str] | None = None) -> Self:
+    def from_json(
+        cls,
+        filepath: os.PathLike | str,
+        distance_function: Callable[[Sequence[Hashable], Sequence[Hashable]], int],
+        key: Callable[[ValueT], str] | None = None,
+        value_type: type[ValueT] = str,
+        node_type: type[NodeT] = BKTreeNode,
+        *,
+        post_add_hook: Callable[[Self, NodeT], None] = lambda tree, node: None,
+        post_already_added_hook: Callable[[Self, NodeT], None] = lambda tree,
+        node: None,
+        post_filter_add_hook: Callable[[Self, NodeT, NodeT], None] = lambda tree,
+        node,
+        filtered_node: None,
+        post_insert_subtree_hook: Callable[[Self, NodeT, NodeT], None] = lambda tree,
+        node,
+        subtree_root: None,
+        post_combine_hook: Callable[[Self, NodeT, NodeT], None] = lambda tree,
+        node,
+        combined_node: None,
+    ) -> Self:
         """Create a BK-Tree from a JSON file.
 
         Args:
             filepath (os.PathLike): The path to the JSON file.
             distance_function (Callable): A function that calculates the distance between two song identifiers
             key (Callable[[ValueT], str] | None): A callable that return the string to be compared.
+            value_type (type[ValueT]): The type of the value to store in the tree.
+            node_type (type[NodeT]): The type of the node to use in the tree.
+            post_add_hook (Callable[[Self, NodeT], None]): A callable that is called after adding a new node to the tree.
+            post_already_added_hook (Callable[[Self, NodeT], None]): A callable that is called when trying to add a node that is
+                already in the tree.
+            post_filter_add_hook (Callable[[Self, NodeT, NodeT], None]): A callable that is called after adding a node to the tree
+            post_insert_subtree_hook (Callable[[Self, NodeT, NodeT], None]): A callable that is called after
+                inserting a subtree to the tree.
+            post_combine_hook (Callable[[Self, NodeT, NodeT], None]): A callable that is called after combining two trees.
 
         Returns:
             CustomBKTree: The BK-Tree.
         """
-        with open(filepath, 'r', encoding="utf-8") as json_file:
+        with open(filepath, "r", encoding="utf-8") as json_file:
             json_data = json.load(json_file)
 
-        return cls.create_tree(json_data, distance_function, key)
+        return cls.create_tree(
+            json_data,
+            distance_function,
+            key,
+            value_type=value_type,
+            node_type=node_type,
+            post_add_hook=post_add_hook,
+            post_already_added_hook=post_already_added_hook,
+            post_filter_add_hook=post_filter_add_hook,
+            post_insert_subtree_hook=post_insert_subtree_hook,
+            post_combine_hook=post_combine_hook,
+        )
 
     @classmethod
-    def create_tree(cls, json_data: Any, distance_function: Callable[[Sequence[Hashable], Sequence[Hashable]], int],
-                    key: Callable[[ValueT], str] | None = None, value_type: type[ValueT] = str,
-                    node_type: type[BKTreeNode] = BKTreeNode) -> Self:
-        tree = cls(distance_function, json_data["max_distance"], key=key, value_type=value_type, node_type=node_type)
+    def create_tree(
+        cls,
+        json_data: Any,
+        distance_function: Callable[[Sequence[Hashable], Sequence[Hashable]], int],
+        key: Callable[[ValueT], str] | None = None,
+        value_type: type[ValueT] = str,
+        node_type: type[NodeT] = BKTreeNode,
+        *,
+        post_add_hook: Callable[[Self, NodeT], None] = lambda tree, node: None,
+        post_already_added_hook: Callable[[Self, NodeT], None] = lambda tree,
+        node: None,
+        post_filter_add_hook: Callable[[Self, NodeT, NodeT], None] = lambda tree,
+        node,
+        filtered_node: None,
+        post_insert_subtree_hook: Callable[[Self, NodeT, NodeT], None] = lambda tree,
+        node,
+        subtree_root: None,
+        post_combine_hook: Callable[[Self, NodeT, NodeT], None] = lambda tree,
+        node,
+        combined_node: None,
+    ) -> Self:
+        tree = cls(
+            distance_function,
+            json_data["max_distance"],
+            key=key,
+            value_type=value_type,
+            node_type=node_type,
+            post_add_hook=post_add_hook,
+            post_already_added_hook=post_already_added_hook,
+            post_filter_add_hook=post_filter_add_hook,
+            post_insert_subtree_hook=post_insert_subtree_hook,
+            post_combine_hook=post_combine_hook,
+        )
         tree.nb_elements = json_data["nb_elements"]
         tree.root = node_type.from_dict(json_data["nodes"], value_type=value_type)
 
@@ -133,7 +216,9 @@ class BKTree[ValueT, NodeT]:
             return NotImplemented
 
         if self.nb_elements != other.nb_elements:
-            logger.debug(f"The number of elements is not the same: {self.nb_elements} - {other.nb_elements}")
+            logger.debug(
+                f"The number of elements is not the same: {self.nb_elements} - {other.nb_elements}"
+            )
             return False
 
         return self.root == other.root
@@ -157,7 +242,7 @@ class BKTree[ValueT, NodeT]:
     def __iter__(self):
         return self.depth_search(self.root)
 
-    def depth_search(self, node: BKTreeNode):
+    def depth_search(self, node: NodeT):
         if node:
             yield node.value
             for child in node.children.values():
@@ -166,7 +251,7 @@ class BKTree[ValueT, NodeT]:
     def __str__(self, level: int = 0):
         return self.root.__str__()
 
-    def contains_node(self, node: BKTreeNode):
+    def contains_node(self, node: NodeT):
         if self.root is None:
             return False
 
@@ -175,7 +260,7 @@ class BKTree[ValueT, NodeT]:
 
         return self._contains_node(node, self.root)
 
-    def _contains_node(self, node: BKTreeNode, comparison_node: BKTreeNode):
+    def _contains_node(self, node: NodeT, comparison_node: NodeT):
         for child_node in comparison_node.children.values():
             if child_node is node:
                 return True
@@ -184,16 +269,21 @@ class BKTree[ValueT, NodeT]:
 
         return False
 
-    def search(self, key_value: str, max_distance: int | None = None, start_node: BKTreeNode | None = None) -> list[BKTreeNode]:
+    def search(
+        self,
+        key_value: str,
+        max_distance: int | None = None,
+        start_node: NodeT | None = None,
+    ) -> list[NodeT]:
         """Searches in the BK-Tree nodes that are within a specified distance from another node.
 
         Args:
             key_value (str): The string value to search for.
             max_distance (int): The maximum allowable distance from the node value.
-            start_node (CustomBKTreeNode | None): The node from which to start searching
+            start_node (NodeT | None): The node from which to start searching
 
         Returns:
-            list[BKTreeNode]: A list of nodes that are within the max_distance from the given value.
+            list[NodeT]: A list of nodes that are within the max_distance from the given value.
         """
         if self.root is None:
             return []
@@ -206,19 +296,19 @@ class BKTree[ValueT, NodeT]:
 
         return self._search(start_node, key_value, max_distance)
 
-    def _search(self, node: BKTreeNode, key_value: str, max_distance: int) -> list[BKTreeNode]:
+    def _search(self, node: NodeT, key_value: str, max_distance: int) -> list[NodeT]:
         """Recursively search in the BK-Tree for nodes that are within a specified distance from another node.
 
         Args:
-            node (BKTreeNode): The current node being compared.
+            node (NodeT): The current node being compared.
             key_value (str): The string value to search for.
             max_distance (int): The maximum allowable distance from the value.
 
         Returns:
-            list[BKTreeNode]: A list of nodes that are within the max_distance from the given value.
+            list[NodeT]: A list of nodes that are within the max_distance from the given value.
         """
         distance = self.distance_function(key_value, self.key(node.value))
-        results: list[BKTreeNode] = []
+        results: list[NodeT] = []
         if distance <= max_distance:
             results.append(node)
 
@@ -230,14 +320,14 @@ class BKTree[ValueT, NodeT]:
 
         return results
 
-    def add(self, value: ValueT) -> tuple[BKTreeNode | None, bool]:
+    def add(self, value: ValueT) -> tuple[NodeT | None, bool]:
         """Add a value to the BK-Tree. If the tree is empty, the value becomes the root.
 
         Args:
             value: The value to add.
 
         Returns:
-            BKTreeNode: The node where the value is stored.
+            NodeT: The node where the value is stored.
             bool: Whether it's a new Node or not.
         """
         string_value = self.key(value)
@@ -245,7 +335,7 @@ class BKTree[ValueT, NodeT]:
 
         if self.root is None:
             logger.debug(f"Setting the root of the tree to {string_value}")
-            self.root = BKTreeNode(value, self.value_type)
+            self.root = self.node_type(value, self.value_type)
             self.nb_elements = 1
             return self.root, True
         else:
@@ -258,29 +348,34 @@ class BKTree[ValueT, NodeT]:
             else:
                 if len(search_result) > 1:
                     logger.warning(
-                        f"Several results for the identifier {string_value}: {[self.key(node.value) for node in search_result]}")
+                        f"Several results for the identifier {string_value}: {[self.key(node.value) for node in search_result]}"
+                    )
                     return search_result[0], False
                 else:
-                    logger.debug(f"{self.key(search_result[0].value)} is similar to {self.key(value)} and is already in the tree.")
+                    logger.debug(
+                        f"{self.key(search_result[0].value)} is similar to {self.key(value)} and is already in the tree."
+                    )
                     self.post_already_added_hook(self, search_result[0])
                     return search_result[0], False
 
-    def _add(self, node: BKTreeNode, value: ValueT) -> BKTreeNode:
+    def _add(self, node: NodeT, value: ValueT) -> NodeT:
         """Recursively adds the song to the correct position in the BK-Tree.
 
         Args:
-            node (BKTreeNode): The current node being compared to the song identifier.
+            node (NodeT): The current node being compared to the song identifier.
             value (T): The value to add to the tree.
 
         Returns:
-            BKTreeNode: The newly created node.
+            NodeT: The newly created node.
         """
-        comparison_distance = self.distance_function(self.key(node.value), self.key(value))
+        comparison_distance = self.distance_function(
+            self.key(node.value), self.key(value)
+        )
 
         if comparison_distance in node.children:
             return self._add(node.children[comparison_distance], value)
         else:
-            new_node = BKTreeNode(value, self.value_type)
+            new_node = self.node_type(value, self.value_type)
             node.children[comparison_distance] = new_node
             self.nb_elements += 1
             self.post_add_hook(self, new_node)
@@ -289,21 +384,21 @@ class BKTree[ValueT, NodeT]:
             )
             return new_node
 
-    def filter(self, filter_fn: Callable[[BKTreeNode], bool]) -> "BKTree":
+    def filter(self, filter_fn: Callable[[NodeT], bool]) -> "BKTree":
         filtered_tree = BKTree(self.distance_function, self.max_distance, key=self.key)
 
         if not self.root:
             return filtered_tree
 
-        def _filter_recursive(node: BKTreeNode):
+        def _filter_recursive(node: NodeT):
             if filter_fn(node):
                 if filtered_tree.root is None:
-                    filtered_tree.root = BKTreeNode(node.value, self.value_type)
+                    filtered_tree.root = self.node_type(node.value, self.value_type)
                     filtered_tree.nb_elements = 1
                     added_node = filtered_tree.root
                 else:
                     added_node = filtered_tree._add(filtered_tree.root, node.value)
-                self.post_filter_add_hook(filtered_tree, added_node)
+                self.post_filter_add_hook(filtered_tree, added_node, node)
 
             for _, child in node.children.items():
                 _filter_recursive(child)
@@ -312,33 +407,45 @@ class BKTree[ValueT, NodeT]:
 
         return filtered_tree
 
-    def insert(self, value: ValueT, start_node: BKTreeNode) -> BKTreeNode | None:
+    def insert(self, value: ValueT, start_node: NodeT) -> NodeT | None:
         """Insert a value in a subtree.
 
         Args:
             value (T): The value to insert
-            start_node (CustomBKTreeNode): The start node of the subtree
+            start_node (NodeT): The start node of the subtree
 
         Returns:
-            CustomBKTreeNode: The newly created node.
+            NodeT: The newly created node.
         """
-        if not isinstance(start_node, BKTreeNode):
-            raise TypeError(f"The start_node must be of type BKTreeNode, not of type {type(start_node)}")
+        if not isinstance(start_node, self.node_type):
+            raise TypeError(
+                f"The start_node must be of type BKTreeNode, not of type {type(start_node)}"
+            )
 
         if not self.contains_node(start_node):
             raise ValueError("The start_node must already be in the BKTree.")
 
-        logger.debug(f"Inserting {value} to the subtree from {self.key(start_node.value)}.")
+        logger.debug(
+            f"Inserting {value} to the subtree from {self.key(start_node.value)}."
+        )
 
         # Search if the identifier or a similar identifier is already in the tree.
-        search_result = self.search(self.key(value), self.max_distance, start_node=start_node)
+        search_result = self.search(
+            self.key(value), self.max_distance, start_node=start_node
+        )
 
         if not search_result:
             # Remove ending content between parentheses or brackets from the identifier.
-            cleaned_identifier = re.sub(r'((\(.+\))|(\[.+]))$', '', self.key(value)).rstrip()
+            cleaned_identifier = re.sub(
+                r"((\(.+\))|(\[.+]))$", "", self.key(value)
+            ).rstrip()
             if cleaned_identifier != self.key(value):
                 # Search if the cleaned_identifier is already in the tree
-                search_result = self.search(cleaned_identifier, max_distance=self.max_distance, start_node=start_node)
+                search_result = self.search(
+                    cleaned_identifier,
+                    max_distance=self.max_distance,
+                    start_node=start_node,
+                )
 
         # If there is still no similar identifier, we add the song to the tree.
         if not search_result:
@@ -348,13 +455,18 @@ class BKTree[ValueT, NodeT]:
         else:
             if len(search_result) > 1:
                 logger.warning(
-                    f"Several results for the identifier {self.key(value)}: {[self.key(node.value) for node in search_result]}")
+                    f"Several results for the identifier {self.key(value)}: {[self.key(node.value) for node in search_result]}"
+                )
                 return None
             else:
-                logger.debug(f"{self.key(search_result[0].value)} is similar to {self.key(value)} and is already in the tree.")
+                logger.debug(
+                    f"{self.key(search_result[0].value)} is similar to {self.key(value)} and is already in the tree."
+                )
                 return search_result[0]
 
-    def search_node_and_parent(self, value: ValueT) -> tuple[BKTreeNode | None, BKTreeNode | None, int | None]:
+    def search_node_and_parent(
+        self, value: ValueT
+    ) -> tuple[NodeT | None, NodeT | None, int | None]:
         """Get the node containing the value and his parent node.
 
         Args:
@@ -368,8 +480,13 @@ class BKTree[ValueT, NodeT]:
         """
         return self._search_node_and_parent(value, self.root)
 
-    def _search_node_and_parent(self, value: ValueT, node: BKTreeNode, parent: BKTreeNode | None = None,
-                                distance_parent: int | None = None) -> tuple[BKTreeNode | None, BKTreeNode | None, int | None]:
+    def _search_node_and_parent(
+        self,
+        value: ValueT,
+        node: NodeT,
+        parent: NodeT | None = None,
+        distance_parent: int | None = None,
+    ) -> tuple[NodeT | None, NodeT | None, int | None]:
         """Get the node containing the value and his parent node.
 
         Args:
@@ -393,14 +510,16 @@ class BKTree[ValueT, NodeT]:
             return node, parent, distance_parent
 
         for dist, child in node.children.items():
-            found_node, found_parent, distance = self._search_node_and_parent(value, child, node, dist)
+            found_node, found_parent, distance = self._search_node_and_parent(
+                value, child, node, dist
+            )
             if found_node:
                 return found_node, found_parent, distance
 
         # If the song isn't found, return None and None.
         return None, None, None
 
-    def get_node_largest_subtree(self, node: BKTreeNode) -> tuple[BKTreeNode | None, int | None]:
+    def get_node_largest_subtree(self, node: NodeT) -> tuple[NodeT | None, int | None]:
         """Getting the node child with the largest size and the distance from the node.
 
         Args:
@@ -419,7 +538,7 @@ class BKTree[ValueT, NodeT]:
 
         largest_subtree_value = 0
         distance_from_parent = 0
-        best_node: BKTreeNode | None = None
+        best_node: NodeT | None = None
 
         for distance, child in node.children.items():
             if (child_length := len(child)) > largest_subtree_value:
@@ -430,7 +549,7 @@ class BKTree[ValueT, NodeT]:
         logger.debug(f"Largest subtree value: {largest_subtree_value}")
         return best_node, distance_from_parent
 
-    def insert_subtree(self, node: BKTreeNode, parent_node: BKTreeNode) -> None:
+    def insert_subtree(self, node: NodeT, parent_node: NodeT) -> None:
         """Insert a subtree in the tree from the parent node passed in parameter.
 
         Args:
@@ -441,7 +560,7 @@ class BKTree[ValueT, NodeT]:
         """
         new_node = self.insert(node.value, parent_node)
         if new_node:
-            self.post_insert_subtree_hook(self, new_node)
+            self.post_insert_subtree_hook(self, new_node, node)
 
         for _, child in node.children.items():
             self.insert_subtree(child, parent_node)
@@ -527,7 +646,7 @@ class BKTree[ValueT, NodeT]:
 
         return values
 
-    def _collect_values(self, node: BKTreeNode, values: list[ValueT]):
+    def _collect_values(self, node: NodeT, values: list[ValueT]):
         values.append(node.value)
 
         for child in node.children.values():
@@ -542,11 +661,11 @@ class BKTree[ValueT, NodeT]:
         if other_tree.root:
             self._combine(other_tree.root)
 
-    def _combine(self, other_tree_node: BKTreeNode):
+    def _combine(self, other_tree_node: NodeT):
         new_node, _ = self.add(other_tree_node.value)
 
         if new_node:
-            self.post_combine_hook(self, new_node)
+            self.post_combine_hook(self, new_node, other_tree_node)
 
         for node in other_tree_node.children.values():
             self._combine(node)
@@ -555,15 +674,19 @@ class BKTree[ValueT, NodeT]:
         """Find the root node with the best efficiency (which has the lowest average distance from other nodes)
 
         Returns:
-            CustomBKTreeNode | Node: The node with the best efficiency or None if the tree is empty.
+            NodeT | Node: The node with the best efficiency or None if the tree is empty.
         """
-        min_avg_distance = float('inf')
+        min_avg_distance = float("inf")
         best_node = None
 
         data = self.collect_values()
 
         for candidate in data:
-            total_distance = sum(self.distance_function(self.key(candidate), self.key(other)) for other in data if other != candidate)
+            total_distance = sum(
+                self.distance_function(self.key(candidate), self.key(other))
+                for other in data
+                if other != candidate
+            )
             avg_distance = total_distance / len(data)
 
             if avg_distance < min_avg_distance:
@@ -587,10 +710,10 @@ class BKTree[ValueT, NodeT]:
         json_data = {
             "nb_elements": self.nb_elements,
             "max_distance": self.max_distance,
-            "nodes": self.root.to_dict() if self.root else {}
+            "nodes": self.root.to_dict() if self.root else {},
         }
 
-        with open(filepath, 'w', encoding="utf-8") as json_file:
+        with open(filepath, "w", encoding="utf-8") as json_file:
             json.dump(json_data, json_file, indent=4, ensure_ascii=False)
 
         return json_data
